#!/usr/bin/env python

# built-in libraries
import os
from time import sleep
# external libraries
import pandas as pd
import numpy as np

def collect_seleced_bstack(folder,buildmodel):
	print('## collect_selected_bstack.py')
	if buildmodel:
		uiname = 'image sets to build model.csv'
		UI = pd.read_csv(os.path.join(folder,uiname))
		setpaths = UI['set location']
		ch1= UI['ch1']
		ch2= UI['ch2']
		c1_stacks = []
		c2_stacks = []
		for setidx, setpath in enumerate(setpaths):
			pickles = [_ for _ in os.listdir(setpath) if _.lower().endswith('pickle')]
			c1_stack = [pd.read_pickle(os.path.join(setpath,pkl)) for pkl in pickles if ch1[setidx] in pkl]
			c2_stack = [pd.read_pickle(os.path.join(setpath,pkl)) for pkl in pickles if ch2[setidx] in pkl]
			c1_stacks = c1_stacks + c1_stack
			c2_stacks = c2_stacks + c2_stack

		df_c1 = pd.concat(c1_stacks,ignore_index=True)
		df_c2 = pd.concat(c2_stacks,ignore_index=True)
		return df_c1,df_c2

	else: 
		raise NameError('collect_seleced_bstack is only for buildmodel')
		return


#!/usr/bin/env python

# built-in libraries
import math
# external libraries
import numpy as np

def reg_bd_svd(bd0=None):
    xc=np.mean(bd0[1])
    yc=np.mean(bd0[0])
    bd0=np.append([bd0[0]]-yc,[bd0[1]]-xc,axis=0) 
    xi=bd0[1]
    yi=bd0[0]
    s=np.sqrt((sum(np.power(xi,2))+sum(np.power(yi,2)))/len(xi)) 
    xi=xi / s
    yi=yi / s
    xiyi=np.append([xi],[yi],axis=0).transpose()
    U, S, rm = np.linalg.svd(xiyi, full_matrices=True) 
    xynew=np.dot(xiyi,rm.transpose())
    xynew=xynew.transpose()
    yc=xynew[1].mean() 
    xc=xynew[0].mean()
    xon=xynew[0] - xc
    yon=xynew[1] - yc
    theta = np.empty(len(yon))
    theta[:] = np.nan
    for i in range(len(yon)):
        theta[i]=math.atan2(yon[i],xon[i])

    cc=np.argwhere(abs(theta) == min(abs(theta)))[0] #error here
    ccid = np.append(range(cc,len(xon)),range(0,cc))
    ccid = ccid.astype(int)
    xon=xon[ccid]
    yon=yon[ccid]
    theta=theta[ccid]
    if theta[4] - theta[0] < 0:
        xon=np.append(xon[-1:0:-1],xon[0])
        yon=np.append(yon[-1:0:-1],yon[0])
    regbd=np.append([yon],[xon],axis=0)
    return regbd, s


#!/usr/bin/env python

# built-in libraries
from time import sleep
import os
import re
# external libraries
import numpy as np 
from PIL import Image

def sum_binary(entries,experiment,progress_bar):
	labeledimfolder = os.path.join(experiment,'segmented images')
	setfolders = [_ for _ in os.listdir(experiment) if (os.path.isdir(os.path.join(experiment,_)) and _ != 'segmented images')]
	setfolderpaths = [os.path.join(experiment,_) for _ in setfolders]
	progress = 0
	for setidx, setid in enumerate(setfolderpaths):
		imfolders = [_ for _ in os.listdir(setid) if os.path.isdir(os.path.join(setid,_))]
		imfolders.sort(key=lambda f: int(filter(str.isdigit,f)))
		imfolderpaths = [os.path.join(setid,_) for _ in imfolders]
		for imidx,imid in enumerate(imfolderpaths):
			masks = [_ for _ in os.listdir(imid) if _.lower().endswith('.tiff')]
			#maskpaths = [os.path.join(imid,_) for _ in masks]
			objectname = list(set([re.split('_+',_)[0] for _ in masks])) # [cell, nuclei]
			## objectname is very limited by re.split. The objectname is something infront of _
			## but this is okay for cell profiler usage.
			for objidx, obj in enumerate(objectname):
				progress = progress + int( (100/len(setfolderpaths)/len(imfolderpaths))*objidx/len(objectname) )
				progress_bar["value"] = progress
				progress_bar.update()
				objs = [_ for _ in masks if obj in _]
				objpaths = [os.path.join(imid,_) for _ in objs]
				for idx, im in enumerate(objpaths):
					addim = np.asarray(Image.open(im))  
					addim = addim.copy()
					addim[addim>0] = idx
					if idx == 0: ims = 0
					ims = addim + ims
				labeledim = Image.fromarray((ims).astype('uint16'))
				labeledsetfolder = os.path.join(labeledimfolder,str(setidx+1))
				if not os.path.isdir(labeledsetfolder):
					os.mkdir(labeledsetfolder)
				dst = os.path.join(labeledsetfolder, obj+'_'+setfolders[setidx]+'_'+imfolders[imidx])
				labeledim.save(dst+'.tiff')





#!/usr/bin/env python

# built-in libraries
import os
from time import sleep
# external libraries
import pandas as pd
import numpy as np

def recordIDX(IDX,buildmodel,folder,cellornuc):
	print('## recordIDX.py')

	if buildmodel: 
		uiname = 'image sets to build model.csv'
	else: 
		uiname = 'image sets to apply model.csv'
	UI = pd.read_csv(os.path.join(folder,uiname))
	setpaths = UI['set location']

	# ledgername = cellornuc + '_object_ledger.csv'
	# picklename = cellornuc + '_boundary_coordinate_stack.pickle'

	if cellornuc == 'cells':
		ledgername = 'Cells_registry.csv'
		picklename = 'Cells_boundary_coordinate_stack.pickle'
	else: 
		ledgername = 'Nuclei_registry.csv'
		picklename = 'Nuclei_boundary_coordinate_stack.pickle'

	for setpath in setpaths:
		obj_ledger = pd.read_csv(os.path.join(setpath,ledgername))
		pkl = pd.read_pickle(os.path.join(setpath,picklename)).values
		pkl = pkl.flatten()
		setlength = len(pkl)
		obj_ledger['IDX']=pd.Series(IDX[0:setlength]) #write
		IDX = np.delete(IDX,range(setlength)) #remove
		obj_ledger.to_csv(os.path.join(setpath,ledgername))




#!/usr/bin/env python

# built-in libraries
from copy import deepcopy
import time
import os
# external libraries
import numpy as np
import pandas as pd
# my files
from bd_resample import *
from reg_bd_svd import *
from reg_bd3 import *

def bdreg_v1(B,N=None,VamModel=None,BuildModel=None):
    print('## bdreg.py')
    start = time.time()
    np.set_printoptions(precision=5,suppress=True)

    if N == None:
        N=50

    if not BuildModel:
        print('applying model')
        N = VamModel['N']
    elif BuildModel:
        print('building model')
        VamModel['N']= N

    plotres=0
    bnreg0=deepcopy(B)
    kll=len(B)
    bdpc=np.zeros([kll,2*N])
    bdpc0=deepcopy(bdpc)
    sc=np.zeros([kll,1])

    for ktt in range(kll): # speed : 3 sec
        bdt=bd_resample((B.loc[ktt]),N)
        B.loc[ktt],sc[ktt]=reg_bd_svd(bdt)
        bdpc0[ktt]=np.append([B[ktt][1]],[B[ktt][0]],axis=1)
    mbdpc0 = [sum(x)/len(x) for x in zip(*bdpc0)]
    bdr0=np.append([mbdpc0[N:]],[mbdpc0[0:N]],axis=0)

    if BuildModel:
        bdrn=deepcopy(bdr0)
        VamModel['bdrn']=bdrn
    else:
        bdrn=VamModel['bdrn']

    outyt=np.zeros([kll,1])
    bnreg=deepcopy(B)

    for ktt in range(kll): # speed : 60 sec
        bnreg[ktt],outyt[ktt]=reg_bd3(bnreg.loc[ktt],bdrn)
        bdpc[ktt]=np.append(bnreg[ktt][1],bnreg[ktt][0])

    return bdpc, bnreg, sc, VamModel

def bdreg_main(df,N=None,VamModel=None,BuildModel=None):
    start = time.time() #record time
    bdpc, bnreg, sc, VamModel=bdreg_v1(df[0],N,VamModel,BuildModel)
    end = time.time()
    print('For bdreg, elapsed time is ' + str(end-start) + 'seconds...')
    return bdpc, bnreg, sc, VamModel





#!/usr/bin/env python

# built-in libraries
import os
import pickle
import random
from time import sleep
# external libraries
import pandas as pd
import numpy as np
# my wrapper
from collect_selected_bstack import *
from recordIDX import *
# my core
from bdreg import *
from pca_bdreg import *
from clusterSM import *

# def main(buildmodel,clnum,folder,entries,modelname,progress_bar=None):
def main(BuildModel,clnum,folder,entries,modelname,progress_bar=None):
	print('## main.py')
	if BuildModel:
		ch1,ch2 = collect_seleced_bstack(folder,BuildModel)

		VamModel = {
		"N":[],
		"bdrn":[],
		"mdd":[],
		"sdd":[],
		"pc":[],
		"latent":[],
		"clnum":[],
		"pcnum":[],
		"mincms":[],
		"testmean":[],
		"teststd":[],
		"boxcoxlambda":[],
		"C":[],
		"Z":[]
		}
		N = None

		ch = 'ch1'
		print(ch)
		progress_bar["value"] = 10
		progress_bar.update()
		bdpc, bnreg, sc, VamModel = bdreg_main(ch1,N,VamModel,BuildModel)
		progress_bar["value"] = 20
		progress_bar.update()
		pc , score, latent, VamModel = pca_bdreg_main(bdpc,VamModel,BuildModel,ch)
		progress_bar["value"] = 30
		progress_bar.update()
		pcnum=None
		IDX,bdsubtype,C,VamModel = cluster_main(folder,modelname,score,pc,bdpc,clnum,pcnum,VamModel,BuildModel,ch)
		progress_bar["value"] = 40
		progress_bar.update()
		if not os.path.exists(os.path.join(folder,'picklejar')):
			os.mkdir(os.path.join(folder,'picklejar'))
		if os.path.exists(os.path.join(os.path.join(folder,'picklejar'),modelname+'_ch1.pickle')):
			f=open(os.path.join(os.path.join(folder,'picklejar'),modelname + str(random.randint(0,100)) +'_ch1.pickle'),'wb')
			pickle.dump(VamModel,f)
			f.close()
		else:
			f=open(os.path.join(os.path.join(folder,'picklejar'),modelname+'_ch1.pickle'),'wb')
			pickle.dump(VamModel,f)
			f.close()
		progress_bar["value"] = 45
		progress_bar.update()
		result = recordIDX(IDX,BuildModel,folder,ch)
		progress_bar["value"] = 50
		progress_bar.update() 

		ch = 'ch2'
		print(ch)
		progress_bar["value"] = 55
		progress_bar.update()
		bdpc, bnreg, sc, VamModel = bdreg_main(ch2,N,VamModel,BuildModel)
		progress_bar["value"] = 60
		progress_bar.update()
		pc , score, latent, VamModel = pca_bdreg_main(bdpc,VamModel,BuildModel,ch)
		progress_bar["value"] = 70
		progress_bar.update()
		pcnum=None
		IDX,bdsubtype,C,VamModel = cluster_main(folder,modelname,score,pc,bdpc,clnum,pcnum,VamModel,BuildModel,ch)
		progress_bar["value"] = 80
		progress_bar.update()

		if os.path.exists(os.path.join(os.path.join(folder,'picklejar'),modelname+'_ch2.pickle')):
			f=open(os.path.join(os.path.join(folder,'picklejar'),modelname + str(random.randint(0,100)) +'_nuclei.pickle'),'wb')
			pickle.dump(VamModel,f)
			f.close()
		else:
			f=open(os.path.join(os.path.join(folder,'picklejar'),modelname+'_ch2.pickle'),'wb')
			pickle.dump(VamModel,f)
			f.close()

		progress_bar["value"] = 90
		progress_bar.update()

		result = recordIDX(IDX,BuildModel,folder,ch)
		progress_bar["value"] = 95
		progress_bar.update() 

	else:
		uiname = 'image sets to apply model.csv'
		UI = pd.read_csv(os.path.join(folder,uiname))
		setpaths = UI['set location']
		ch1ui= UI['ch1']
		ch2ui= UI['ch2']    
		condition = UI['condition']

		for setidx, setpath in enumerate(setpaths):
			pickles = [_ for _ in os.listdir(setpath) if _.lower().endswith('pickle')]

			c1_stack = [pd.read_pickle(os.path.join(setpath,pkl)) for pkl in pickles if ch1ui[setidx] in pkl]
			c2_stack = [pd.read_pickle(os.path.join(setpath,pkl)) for pkl in pickles if ch2ui[setidx] in pkl]
			
			ch1 = pd.concat(c1_stack,ignore_index=True)
			ch2 = pd.concat(c2_stack,ignore_index=True)

			progress_bar["value"] = 10
			progress_bar.update()

			try:
				f=open(os.path.join(os.path.join(folder,'picklejar'), modelname +'_ch1.pickle'),'r')
			except:
				entries['Status'].delete(0,END) #global name END is not defined
				entries['Status'].insert(0,'the model does not exist. please replace model name to the one you built')
			VamModel = pickle.load(f)

			N = VamModel['N'] 
			ch = 'ch1'
			print(ch)
			progress_bar["value"] = 20
			progress_bar.update()
			bdpc_new, bnreg_new, sc_new, VamModel = bdreg_main(ch1,N,VamModel,BuildModel)
			progress_bar["value"] = 30
			progress_bar.update()
			pc_new, score_new, latent_new, VamModel = pca_bdreg_main(bdpc_new,VamModel,BuildModel,ch)
			progress_bar["value"] = 40
			progress_bar.update()
			if clnum != VamModel['clnum']:
				raise NameError('Number of eigen-shape should remain same. To change the number, rebuild the model with the new number first')
			clnum=VamModel['clnum']
			pcnum=VamModel['pcnum']
			#pc_new goes in for sake of placing, but pc from the model is used in cluster_main
			IDX_new,bdsubtype_new,C_new,VamModel = cluster_main(folder,modelname,score_new,pc_new,bdpc_new,clnum,pcnum,VamModel,BuildModel,ch,condition[setidx])
			progress_bar["value"] = 50
			progress_bar.update()

			result = recordIDX(IDX_new,BuildModel,folder,ch)

			try:
				f=open(os.path.join(os.path.join(folder,'picklejar'), modelname +'_ch2.pickle'),'r')
			except:
				print('error')
				# entries['Status'].delete(0,END)
				# entries['Status'].insert(0,'the model does not exist. please replace model name to the one you built')
			VamModel = pickle.load(f)

			N = VamModel['N'] 
			ch = 'ch2'
			print(ch)
			progress_bar["value"] = 60
			progress_bar.update()
			bdpc_new, bnreg_new, sc_new, VamModel = bdreg_main(ch2,N,VamModel,BuildModel)
			progress_bar["value"] = 70
			progress_bar.update()
			pc_new, score_new, latent_new, VamModel = pca_bdreg_main(bdpc_new,VamModel,BuildModel,ch)
			progress_bar["value"] = 80
			progress_bar.update()
			clnum=VamModel['clnum']
			pcnum=VamModel['pcnum']
			#pc_new goes in for sake of placing, but pc from the model is used in cluster_main
			IDX_new,bdsubtype_new,C_new,VamModel = cluster_main(folder,modelname,score_new,pc_new,bdpc_new,clnum,pcnum,VamModel,BuildModel,ch,condition[setidx])
			progress_bar["value"] = 90
			progress_bar.update()

			result = recordIDX(IDX_new,BuildModel,folder,ch)



#!/usr/bin/env python

# built-in libraries
from __future__ import division
from copy import deepcopy
import time
# external libraries
import numpy as np
import pandas as pd
import matplotlib as mpl
mpl.use("TkAgg")
import matplotlib.pyplot as plt
# my files
from PCA import *

def pca_bdreg_v1(bdpc,VamModel,BuildModel,ch):
    print('## pca_bdreg.py')
    np.set_printoptions(precision=5,suppress=True)

    Nuu = int(round(len(bdpc.T[0])))
    Nbb = int(round(len(bdpc[0])/2))
    pcnum0 = 12
    dstp = 3
    bdpct = deepcopy(bdpc)

    if BuildModel:
        mmx = np.ones((Nuu,1)) * np.mean(bdpct,axis=0)
    else: 
        mmx = np.ones((Nuu,1)) * VamModel['mdd']   
    smx = np.ones(bdpct.shape)
    test = np.divide((bdpct-mmx),smx)
    if BuildModel:
        pc,score,latent=PCA(test)
        score = np.dot(test,pc)
    else: 
        latent = VamModel['latent']
        pc = VamModel['pc']
        score = np.dot(test,pc)

    mdd = mmx[0]
    sdd = smx[0]
    mss = np.mean(score,axis=0) 
    sss = np.std(score,axis=0) 
    cnum = 1

    ce1 = score[cnum-1] 

    xr = bdpct[cnum-1][0:Nbb]
    yr = bdpct[cnum-1][Nbb:]

    VamModel['mdd']=mdd
    VamModel['sdd']=sdd
    VamModel['pc']=pc
    VamModel['latent']=latent

    plt.figure(1)
    plt.clf()

    #check I: check the shape of principle component; 
    if True:
        plt.figure(221)
        plt.clf()
        for pcnum in range(10):
            dxx = 0.7
            xx = pc.T[pcnum][0:Nbb]
            yy = pc.T[pcnum][Nbb:]
            plt.plot(xx+(pcnum+1)*dxx,yy,'b-',linewidth=2.0)
            plt.plot(-xx+(pcnum+1)*dxx,-yy-dxx,'r-',linewidth=2.0)
        plt.axis('equal')
        plt.axis('off')


    #check II: the mean +- std of PC mean
    if True:
        cmap = plt.cm.jet
        vmax = int(11)
        norm = mpl.colors.Normalize(vmin=0,vmax=vmax)
        cid = plt.cm.ScalarMappable(norm=norm,cmap=cmap)

        offx,offy=np.meshgrid(range(pcnum0),[0])
        offx=np.multiply((offx+1),dstp)[0]
        offy=np.multiply(-(offy+1),dstp)[0]

        for kkk in range(pcnum0):
            count=1
            for ks in np.linspace(-10,10,11):
                pnn = np.zeros([len(pc[0])])
                for k in range(Nbb):
                    if k == kkk:
                        pnn=np.add(pnn,pc.T[k]*(mss[k]+ks*sss[k])) #pc[k] has sign issue again
                    else:
                        pnn=pnn+pc.T[k]*mss[k]
                pnn = np.multiply(pnn,sdd.T) + mdd.T
                xx = pnn[0:Nbb]
                yy = pnn[Nbb:]
                xx = np.append(xx,xx[0])
                yy = np.append(yy,yy[0])
                plt.figure(1)
                plt.plot(xx+offx[kkk],yy+offy[kkk],'-',color=cid.to_rgba(count))
                count = count + 1
        plt.axis('equal')
    # check III- reconstruct the cell shape using principle
    if True:
        pnn=np.zeros(len(pc[0]))
        for k in range(pcnum0):
            pnn = pnn + pc.T[k] * ce1[k]
        pnn=np.multiply(pnn,sdd.T) + mdd.T
        xx=pnn[0:Nbb]
        yy=pnn[Nbb:]
        xx=np.append(xx,xx[0])
        yy=np.append(yy,yy[0])
        plt.figure(33)
        plt.plot(np.append(xr,xr[0]),np.append(yr,yr[0]),'r-',xx,yy,'b-')
        plt.axis('equal')
        plt.axis('off')
    count=1
    pnn=np.zeros(len(pc[0]))
    for k in range(pcnum0):
        pnn=pnn + np.multiply(pc.T[k],mss[k])
    pnn=np.multiply(pnn,sdd.T) + mdd.T
    xx=pnn[0:Nbb]
    yy=pnn[Nbb:]
    xx=np.append(xx,xx[0])
    yy=np.append(yy,yy[0])
    plt.figure(1311)
    plt.plot(xx,yy,'b-',linewidth=2.0)
    plt.axis('equal')
    plt.axis('off')
    #if ch == 'ch2': plt.show()
    return pc,score,latent,VamModel

def pca_bdreg_main(bdpc,VamModel,BuildModel,ch):
    start = time.time()
    pc,score,latent,VamModel=pca_bdreg_v1(bdpc,VamModel,BuildModel,ch)
    end = time.time()
    print ('For PCA, elapsed time is ' + str(end-start) + 'seconds...')
    return pc, score, latent, VamModel
